/* A daemon which communicates with auvoxware server

When a client requests a clip be played, the clip's filename
is cached, its samples are loaded into a bucket at the auvoxware
server, and that bucket is played.  Future requests are looked up
in the local cache before being played.

Brian Sellden
*/

#include <iostream.h>
#include <stdlib.h> 
#include <errno.h> 
#include <string.h> 
#include <sys/types.h> 
#include <netinet/in.h> 
#include <sys/socket.h> 
#include <sys/wait.h> 

#include "mNasD.h"

mNasD::mNasD()
{
  initialized = 0;
  monoDevice = AuNone;
  stereoDevice = AuNone;
  currentID = 1;
  retBuf[SUCCESS] = (char*)malloc(strlen("Ok") +1);
  retBuf[FAILURE] = (char*)malloc(strlen("Fail") +1);
}

mNasD::mNasD(const char *servName, int new_fd)
{
  initialized = 0;
  monoDevice = AuNone;
  stereoDevice = AuNone;
  currentID = 1;
  clientFD = new_fd;
initialize(servName);
}

mNasD::~mNasD()
{
  SoundStructPtr ss = NULL;
  mDictValPtrIterator<int, SoundStruct> idIter(*idDict);
  mString *fn=NULL;

  /* Empty these first */
  for(idIter.reset(); idIter.isValid(); idIter.next()) {
    if(ss=idDict->lookup(idIter.current()))
      AuDestroyFlow(aud, ss->flow, NULL);
  }

  /* Both elements of the key/value pair need to be deleted */
  while(soundDict->size() > 0) {
    fn = soundDict->pickKey();
    soundDict->del_item(fn); /* deletes both key and value */
  }

    delete idDict;
    delete soundDict;
  free(retBuf[SUCCESS]);
  free(retBuf[FAILURE]);
  AuCloseServer(aud);
}

int mNasD::ready()
{
  return initialized;
}

int mNasD::initialize(const char *servName)
{
  initialized = 0;
  int i;

  if (!(aud = AuOpenServer(servName, 0, NULL, 0, NULL, NULL))) {
      /* Can't connect to server */
    cerr<<"mNasD::initialize: error - cannot connect to auvoxware server"<<endl;
    return initialized;
  }

    /* look for an output device */
    for (i = 0; i < AuServerNumDevices(aud); i++)
	if ((AuDeviceKind(AuServerDevice(aud, i)) ==
	     AuComponentKindPhysicalOutput)) {
	  if(AuDeviceNumTracks(AuServerDevice(aud, i)) == 1)
	    monoDevice = AuDeviceIdentifier(AuServerDevice(aud, i));

	  if(AuDeviceNumTracks(AuServerDevice(aud, i)) == 2)
	    stereoDevice = AuDeviceIdentifier(AuServerDevice(aud, i));
	}

    if ((monoDevice == AuNone) && (stereoDevice == AuNone)) {
      cerr<<"mNasD::initialize: error - couldn't find export device"<<endl;
      return initialized;
    }

  idDict = new mDictValPtr<int, SoundStruct>;
  soundDict = new mDictPtrPtr<mString, SoundStruct>;

  initialized = 1;

  return initialized;
}

void mNasD::serveClient(int new_fd)
{
  fd_set rfds;
  struct timeval tv;
  int retval, numBytes=-1;
  char result[16];

  if(new_fd <0) {
    /* invalid fd */
    cerr<<"mNasD::serveClient: error - invalid fd"<<endl;
    return;
}
  clientFD = new_fd;

  while(1) {
  /* Watch clientFD to see when it has input. */
  FD_ZERO(&rfds);
  FD_SET(clientFD, &rfds);

  /* Playing with different timeout values.
     It will end up being either about 1/5 seconds, or a very
    long period. */
  tv.tv_sec = 0;
  tv.tv_usec = 100000;

  retval = select(clientFD+1, &rfds, NULL, NULL, &tv);
  /* Don't rely on the value of tv now! */

  if(retval) {
            if((numBytes=recv(clientFD, buf, MAXDATASIZE-1, 0)) > 0) {
	  buf[numBytes] = '\n';
	  AuHandleEvents(aud);
	  /* Handle the request and return the outcome */
	  handleRequest(numBytes, result);
	  send(clientFD, result, strlen(result)+1, 0);

	  AuHandleEvents(aud);
	  if(done == 1) break;
	    }
  } /* if retval */
  } /* while */

}

int mNasD::handleRequest(int numBytes, char *result)
{
  NasRequest req;
  char *myBuf = buf, tmpBuf[3];
  int retVal;

  if(numBytes < 2) {
    cerr<<"mNasD::handleRequest: error - numBytes < 2"<<endl;
    sprintf(result, "%d\n", -1);
    return -1;
  }

  strncpy(tmpBuf, buf, 3);
  tmpBuf[2] = '\n';
  req = (NasRequest)atoi(tmpBuf);

  myBuf+=2;
  numBytes-=2;

  switch(req) {
  case NasLoadSound: {
    int id;

    if(numBytes < 2) {
      cerr<<"mNasD::handleRequest: error - numBytes < 2"<<endl;
      retVal=-1;
      break;
    }

    strncpy(tmpBuf, myBuf, 3);
    tmpBuf[2] = '\n';
    id = atoi(tmpBuf);
    myBuf+=2;
    numBytes-=2;

#if DEBUG_AUDIO
    cout<<"mNasD::handleRequest: request is NasLoadSound"<<endl;
#endif
    retVal=handleLoadRequest(id, myBuf);
    break;
  }
  case NasStartSoundDirect: {
#if DEBUG_AUDIO
    cout<<"mNasD::handleRequest: request is NasStartSoundDirect"<<endl;
#endif
    retVal = handleStartRequest(myBuf);
    break;
  }
  case NasStartSoundCached: {
    int id;
    int vol;

    strncpy(tmpBuf, myBuf, 3);
    tmpBuf[2] = '\n';
    id = atoi(tmpBuf);
    myBuf+=2;

    strncpy(tmpBuf, myBuf, 3);
    tmpBuf[2] = '\n';
    vol = atoi(tmpBuf);

#if DEBUG_AUDIO
    cout<<"mNasD::handleRequest: request is NasStartSoundCached"<<endl;
#endif
    retVal = handleStartRequest(id, vol);
    break;
  }
  case NasStopSoundCached: {
    int id;

    strncpy(tmpBuf, myBuf, 3);
    tmpBuf[2] = '\n';
    id = atoi(tmpBuf);

#if DEBUG_AUDIO
    cout<<"mNasD::handleRequest: request is NasStopSoundCached"<<endl;
#endif
    retVal = handleStopRequest(id);
    break;
  }
  case NasStopSoundDirect: {
#if DEBUG_AUDIO
    cout<<"mNasD::handleRequest: request is NasStopSoundDirect"<<endl;
#endif
    retVal = handleStopRequest(myBuf);
    break;
  }
  case NasStopAllSounds:
#if DEBUG_AUDIO
    cout<<"mNasD::handleRequest: request is NasStopAllSounds"<<endl;
#endif
    retVal=handleStopAllRequest();
    break;
  case NasQuit:
#if DEBUG_AUDIO
    cout<<"mNasD::handleRequest: request is NasQuit"<<endl;
#endif
    handleQuitRequest();
    retVal=0;
    break;
  default:
    cout<<"mNasD::handleRequest: unknown request - "<<req<<endl;
    retVal=-1;
    break;
  }

  sprintf(result, "%d\n", retVal);
  return retVal;
}

  int mNasD::handleLoadRequest(int id, const char *fileName)
    {
    AuBucketID      bucketID;
    AuBucketAttributes *ba;
    SoundStructPtr ss = NULL;
    AuElement       *elements;
    AuUint32        samples = AuUnlimitedSamples;
    AuInt32         offset = 0;
    AuStatus status;

      if((id < 0) || (id > 99)) {
	cerr<<"mNasD::handleLoadRequest: error - id out of range (0-99)"<<endl;
	return 1;
      }

      if(fileName == NULL) {
	cerr<<"mNasD::handleLoadRequest: error - fileName == NULL"<<endl;
	return 1;
      }

      ss = new SoundStruct;

      if (!(ss->flow = AuCreateFlow(aud, NULL))) {
	cerr<<"mNasD::handleLoadRequest: error - couldn't create flow"<<endl;
	return 1;
      }

#if DEBUG_AUDIO
  cout<<" - loading "<<fileName<<endl;
#endif
    if (!(bucketID = AuSoundCreateBucketFromFile(aud, fileName, AuAccessAllMasks, &ba,
						 &status))) {
      AuDestroyFlow(aud, ss->flow, NULL);
      cerr<<"mNasD::handleLoadRequest: error - couldnt make bucket from "<<fileName<<endl;
      switch(status) {
      case AuBadValue :
	cerr<<" - error was AuBadValue"<<endl;
	break;
      case AuBadLength:
	cerr<<" - error was AuBadLength"<<endl;
	break;
      case AuBadAlloc:
	cerr<<" - error was AuBadAlloc"<<endl;
	break;
      default:
	cerr<<" - unknown error"<<endl;
      }
      return 1;
    }

    /* If we got this far, we have at least one device */
    if(AuBucketNumTracks(ba) == 2) {
      if(stereoDevice != AuNone)
	ss->device = stereoDevice;
      else
	ss->device = monoDevice;
    }
    else {/* sound has only 1 track */
      if(monoDevice != AuNone)
	ss->device = monoDevice;
      else
	ss->device = stereoDevice;
    }

    elements = (AuElement*)malloc(2*sizeof(AuElement));

    AuMakeElementImportBucket(&elements[0], AuBucketSampleRate(ba), bucketID,
			      AuBucketNumSamples(ba), 0, 0, NULL);
    AuMakeElementExportDevice(&elements[1], 0, ss->device,
			      AuBucketSampleRate(ba),
			      AuUnlimitedSamples, 0, NULL);

    AuSetElements(aud, ss->flow, AuTrue, 2, elements, NULL);
    AuFreeBucketAttributes(aud, 1, ba);
    ss->id = id;
    idDict->insert(id, ss);

#if DEBUG_AUDIO
  cout<<" - loaded "<<fileName<<endl;
#endif

      return 0;
    }

void mNasD::handleQuitRequest()
{
  done = 1;
}

int mNasD::handleStartRequest(int id, int vol)
{
  SoundStructPtr ss = NULL;
  AuDeviceAttributes da;

  if(id == 0) /* Universal empty sample */
    return 0;

  if(vol < 0)
    return -1;

  if(ss=idDict->lookup(id)) {
    /* set the volume */
    AuDeviceGain(&da) = AuFixedPointFromSum(vol, 0);
    AuSetDeviceAttributes(aud, ss->device, AuCompDeviceGainMask, &da, NULL);
    AuStartFlow(aud, ss->flow, NULL);
    return 0;
  }
  else {
    cerr<<"mNasD::handleStartRequest: warning - bad request, id is "<<id<<endl;
    return 1;
  }
}

/*
This method checks to see if the clip is cached.  If so, it
plays it.  If not, it loads the clip, caches it, and then
plays it.
*/
int mNasD::handleStartRequest(const char *fileName)
{
  mString *fn = NULL;
  SoundStructPtr ss = NULL;
  int vol=-1;
  char volBuf[3];

  if(strlen(fileName) < 3) {
    cerr<<"mNasD::handleStartRequest: error - data too short"<<endl;
    return -1;
  }

strncpy(volBuf, fileName, 3);
 volBuf[2] = '\n';
 vol = atoi(volBuf);
 if(vol < 0) {
   cerr<<"mNasD::handleStartRequest: error - bad value for volume"<<endl;
   return -1;
 }

 fileName+=2;
  fn = new mString(fileName);
  AuDeviceAttributes da;

  /* Check to see if this file has already been played.
  If so, play the cached flow.  Otherwise, create a flow, cache
  it and play it.
  */
  if(soundDict->lookup(fn) != NULL) {
    ss = soundDict->lookup(fn);

    /* set the volume */
    AuDeviceGain(&da) = AuFixedPointFromSum(vol, 0);
    AuSetDeviceAttributes(aud, ss->device, AuCompDeviceGainMask, &da, NULL);
    AuStartFlow(aud, ss->flow, NULL);
    delete fn;
    return ss->id;
  } else {

    ss = new SoundStruct;
    ss->flow=AuNone;
    if(createFlowFromFile(fileName, ss)) {
      cerr<<"mNasD::handleStartRequest: error - couldn't create flow from "<<*fn<<endl;
      delete fn;
      delete ss;
      /* Returning 0 instead of -1, because -1
    causes the code in audio.tcl to abort execution.  This was
    leading to objects not getting spoken.  A flowID of 0
    is now the universal blank sample */
      return 0;
    }

    /* Successfully created flow - fill in the dicts and start
  the flow.
    */
    currentID++;
    ss->id = currentID;
    idDict->insert(ss->id, ss);
    soundDict->insert(fn, ss);

    /* set the volume */
    AuDeviceGain(&da) = AuFixedPointFromSum(vol, 0);
    AuSetDeviceAttributes(aud, ss->device, AuCompDeviceGainMask, &da, NULL);
    AuStartFlow(aud, ss->flow, NULL);
    return ss->id;
  }
}

int mNasD::handleStopRequest(int id)
{
  AuFlowID flow;
  SoundStructPtr ss = NULL;
  SoundStructPtr ss2 = NULL;

#if DEBUG_AUDIO
    cout<<"mNasD::handleStopRequest: stopping clip "<<id<<endl;
#endif

  if(ss2=idDict->lookup(id)) {
    mDictPtrPtrIterator<mString, SoundStruct> iter(*soundDict);
    for(iter.reset(); iter.isValid(); iter.next()) {
      ss = soundDict->lookup(iter.current());
      if(ss->id == ss2->id) {
#if DEBUG_AUDIO
	cout<<"mNasD::handleStopRequest"
	    <<"stopping clip "<<id<<": "<<iter.current()<<endl;
#endif
	break;
      }
    }

    AuStopFlow(aud, ss2->flow, NULL);
    return 0;
  }

  return -1;
}

int mNasD::handleStopRequest(const char *fileName)
{
  mString *fn= NULL;
  SoundStructPtr ss = NULL;

  if(fileName == NULL) {
    cerr<<"mNasD::handleStopRequest: error - fileName == NULL"<<endl;
    return -1;
  }

  fn = new mString(fileName);
  if(soundDict->lookup(fn) != NULL) {
    ss = soundDict->lookup(fn);
    AuStopFlow(aud, ss->flow, NULL);
    return 0;
  }

  /* Flow did not exist in soundDict */
  return -1;
}

 int mNasD::createFlowFromFile(const char *fileName, SoundStruct *ss)
   {
    AuBucketID      bucketID;
    AuBucketAttributes *ba;
    AuElement       elements[2];
    AuUint32        samples = AuUnlimitedSamples;
    AuInt32         offset = 0;
    AuStatus status;

    if(fileName == NULL) {
      cerr<<"mNasD::createFlowFromFile: error - fileName == NULL"<<endl;
      return 1;
    }

      if(ss == NULL) {
	cerr<<"mNasD::createFlowFromFile: error - ss == NULL"<<endl;
	return 1;
      }

      if (!(ss->flow = AuCreateFlow(aud, NULL))) {
	cerr<<"mNasD::createFlowFromFile: error - couldn't create flow"<<endl;
	return 1;
      }

#if DEBUG_AUDIO
  cout<<" - loading "<<fileName<<endl;
#endif

    if (!(bucketID = AuSoundCreateBucketFromFile(aud, fileName, AuAccessAllMasks, &ba,
						 &status))) {
      AuDestroyFlow(aud, ss->flow, NULL);
      ss->flow=AuNone;
      cerr<<"mNasD::createFlowFromFile: error - couldn't make bucket from "<<fileName<<endl;
      switch(status) {
      case AuBadValue :
	cerr<<" - error was AuBadValue"<<endl;
	break;
      case AuBadLength:
	cerr<<" - error was AuBadLength"<<endl;
	break;
      case AuBadAlloc:
	cerr<<" - error was AuBadAlloc"<<endl;
	break;
      default:
	cerr<<" - unknown error"<<endl;
      }
      return 1;
    }

    /* If we got this far, we have at least one device */
    if(AuBucketNumTracks(ba) == 2) {
      if(stereoDevice != AuNone)
	ss->device = stereoDevice;
      else
	ss->device = monoDevice;
    }
    else {/* sound has only 1 track */
      if(monoDevice != AuNone)
	ss->device = monoDevice;
      else
	ss->device = stereoDevice;
    }

    AuMakeElementImportBucket(&elements[0], AuBucketSampleRate(ba), bucketID,
			      AuBucketNumSamples(ba), 0, 0, NULL);
    AuMakeElementExportDevice(&elements[1], 0, ss->device,
			      AuBucketSampleRate(ba),
			      AuUnlimitedSamples, 0, NULL);

    AuSetElements(aud, ss->flow, AuTrue, 2, elements, NULL);
    AuFreeBucketAttributes(aud, 1, ba);

#if DEBUG_AUDIO
  cout<<" - loaded "<<fileName<<endl;
#endif

      return 0;
   }

int
mNasD::handleStopAllRequest()
{
  mDictValPtrIterator<int, SoundStruct> iter(*idDict);
  SoundStructPtr ss = NULL;

  for(iter.reset(); iter.isValid(); iter.next()) {
    ss = idDict->lookup(iter.current());
    AuStopFlow(aud, ss->flow, NULL);
  }
  return 0;
}
